import asyncio
import json
import logging
import socket
import time
from collections import defaultdict
from typing import DefaultDict, NoReturn, Optional, Type

import nats.errors
import requests
from nats.aio.client import Client as NATSClient
from nats.aio.msg import Msg
from pydantic import BaseModel, Field

from powerflex_monitoring.format_exception import format_exception
from powerflex_monitoring.monitoring import MonitoringServer

BaseSettings: Type[BaseModel]
try:
    # type: ignore
    from pydantic_settings import BaseSettings
except ImportError:
    # pylint: disable=ungrouped-imports
    from pydantic import BaseSettings  # type: ignore

logger = logging.getLogger(__name__)
HOSTNAME_TEMPLATE_VAR = "{hostname}"


class NatsHealthCheckerConfig(BaseSettings):  # type: ignore
    NATS_SERVICE_HEALTH_CHECK_ADDRESS: Optional[str] = Field(
        description="NATS service health check HTTP address."
    )
    NATS_SERVICE_HEALTH_CHECK_TIMEOUT: float = Field(
        description="How long to wait for the NATS health endpoint to respond to this service",
        default=5,
    )

    NATS_HEALTH_CHECK_SUBJECT_TEMPLATE: str = Field(
        description="NATS subject on which this service will subscribe to and periodically send requests. "
        f"The string {HOSTNAME_TEMPLATE_VAR} is replaced with the service's hostname. "
        "The hostname is used in this NATS subject to ensure it is unique across all containers. "
        "In docker-compose and kubernetes, the hostname is typically container-specific. ",
        default=f"nats_health_check.echo.{HOSTNAME_TEMPLATE_VAR}",
    )

    NATS_HEALTH_CHECK_UNHEALTHY_TIMEOUT_SEC: float = Field(
        description="Mark the service as unhealthy after determining that NATS is unhealthy for this period of time. "
        "Recommend making it less than NATS_HEALTH_CHECK_REQUEST_PERIOD_SEC.",
        default=60 * 4,
    )

    NATS_HEALTH_CHECK_REQUEST_TIMEOUT_SEC: float = Field(
        description="Timeout for the requests we periodically send to test the NATS connection.",
        default=5,
    )

    NATS_HEALTH_CHECK_REQUEST_PERIOD_SEC: float = Field(
        description="How often to periodically send a NATS request to test the NATS connection.",
        default=15,
    )


class NatsHealthChecker:
    def __init__(self, nats_conn: NATSClient, monitoring_server: MonitoringServer):
        self.config = NatsHealthCheckerConfig(
            # type:ignore
        )

        logger.info(
            "Configuration for NATS health check", extra={"config": self.config.dict()}
        )

        self.nats_health_check_subject = (
            # pylint: disable=no-member
            self.config.NATS_HEALTH_CHECK_SUBJECT_TEMPLATE.replace(
                HOSTNAME_TEMPLATE_VAR,
                socket.gethostname(),
            )
        )
        self.nats_conn = nats_conn
        self.monitoring_server = monitoring_server
        self.nats_request_timeout = 10
        self.time_since_nats_unhealthy: DefaultDict[str, Optional[float]] = defaultdict(
            lambda: None
        )

    async def _async_setup(self) -> None:
        logger.info(
            "NATS health checker subscribing to NATS subject %s to run periodic NATS health check.",
            self.nats_health_check_subject,
        )
        await self.nats_conn.subscribe(
            self.nats_health_check_subject, cb=self._echo_requests
        )

    def on_unhealthy(self, health_check_name: str, cause: str) -> None:
        """Check if a NATS health check has been unhealthy for too long.

        If so, then mark the service as unhealthy.
        """
        self.on_nats_health_status(
            healthy=False, health_check_name=health_check_name, cause=cause
        )

        time_since = self.time_since_nats_unhealthy[health_check_name]
        unhealthy_for_too_long = (
            time_since is not None
            and time.time() - time_since
            >= self.config.NATS_HEALTH_CHECK_UNHEALTHY_TIMEOUT_SEC
        )
        if unhealthy_for_too_long:
            self.monitoring_server.health_check.make_unhealthy(cause)

    def on_nats_health_status(
        self, healthy: bool, health_check_name: str, cause: str = "NATS is healthy"
    ) -> None:
        """Call this whenever a result for a NATS health check is found."""
        if healthy is True:
            logger.debug(
                "NATS health check %s indicates NATS is healthy", health_check_name
            )
            self.time_since_nats_unhealthy[health_check_name] = None

        if healthy is False:
            self.monitoring_server.ready_check.set_ready(
                ready=False, cause=cause, readiness_identifier="nats-identifier"
            )

            if self.time_since_nats_unhealthy[health_check_name] is None:
                logger.info(
                    "NATS health check %s indicates NATS is currently unhealthy",
                    health_check_name,
                    extra={"cause": cause},
                )
                self.time_since_nats_unhealthy[health_check_name] = time.time()

        all_health_checks_pass = all(
            check_failure_time is None
            for check_failure_time in self.time_since_nats_unhealthy.values()
        )
        if not self.monitoring_server.ready_check.is_ready and all_health_checks_pass:
            self.monitoring_server.ready_check.set_ready(
                ready=True,
                cause="NATS is healthy again",
                readiness_identifier="nats-identifier",
            )

    async def _echo_requests(self, nats_msg: Msg) -> None:
        if nats_msg.reply:
            await self.nats_conn.publish(nats_msg.reply, nats_msg.data)

    async def check_echo_nats_request(self) -> None:
        """Make a NATS request to a service-unique subject to check NATS health."""
        health_check_name = self.check_echo_nats_request.__name__
        timeout = self.config.NATS_HEALTH_CHECK_REQUEST_TIMEOUT_SEC
        test_payload = {"timestamp": time.time()}
        try:
            response = await self.nats_conn.request(
                self.nats_health_check_subject,
                json.dumps(test_payload).encode(),
                timeout=timeout,
            )
        except (nats.errors.Error, nats.errors.TimeoutError) as exc:
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause=f"NATS request to test subject failed after timeout={timeout}. Exception: "
                + format_exception(exc),
            )
            return
        except (
            Exception  # pylint: disable=broad-except
        ) as unexpected_nats_request_exception:
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause="Unexpected exception when checking NATS service request. Exception: "
                + format_exception(unexpected_nats_request_exception),
            )
            return
        test_response = json.loads(response.data)
        if test_payload != test_response:
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause="NATS request to test subject returned incorrect data. "
                f"Expected {test_payload}. "
                f"Received {test_response}.",
            )
        else:
            self.on_nats_health_status(
                healthy=True, health_check_name=health_check_name
            )

    async def check_nats_service_health(self) -> None:
        """Check the NATS service's health endpoint to determine if it is healthy."""
        health_check_name = self.check_nats_service_health.__name__
        address = self.config.NATS_SERVICE_HEALTH_CHECK_ADDRESS
        # TODO
        # We will have to switch to aiohttp when  PR #58 is merged
        # https://github.com/edf-re/powerflex_edge_storage_interface/pull/58
        if not address:
            logger.warning(
                "Do not call check_nats_service_health without setting NATS_SERVICE_HEALTH_CHECK_ADDRESS"
            )
            return
        try:
            response = requests.get(
                address, timeout=self.config.NATS_SERVICE_HEALTH_CHECK_TIMEOUT
            )
        except requests.exceptions.ConnectionError as exc:
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause=f"Could not connect to NATS service health check address={address}. Exception: "
                + format_exception(exc),
            )
            return
        except (
            Exception  # pylint: disable=broad-except
        ) as unexpected_nats_health_exception:
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause="Unexpected exception when checking NATS service health. Exception: "
                + format_exception(unexpected_nats_health_exception),
            )
            return

        status = response.status_code

        if status != 200:
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause=f"NATS service health check returned non-200 status code {status}, address={address}",
            )
        else:
            self.on_nats_health_status(
                healthy=True, health_check_name=health_check_name
            )

    async def periodically_check_nats_health(self) -> NoReturn:
        """Periodically check the NATS health by running the following checks.

        1. Check if a test subject only used by this service responds to a NATS request
        2. Check if the NATS service's health check is bad

        If any health check keeps failing for longer than
        NATS_HEALTH_CHECK_UNHEALTHY_TIMEOUT_SEC,
        then mark the service as unhealth.

        If any health check fails at all, mark the service as not ready.

        Mark the service as ready if all health checks pass.
        """
        await self._async_setup()

        logger.info(
            "Starting NATS health checker(s): check_echo_nats_request%s",
            (
                " check_nats_service_health"
                if self.config.NATS_SERVICE_HEALTH_CHECK_ADDRESS
                else ""
            ),
        )
        while True:
            tasks = [self.check_echo_nats_request()]
            if self.config.NATS_SERVICE_HEALTH_CHECK_ADDRESS:
                tasks.append(self.check_nats_service_health())
            await asyncio.gather(*tasks)
            await asyncio.sleep(self.config.NATS_HEALTH_CHECK_REQUEST_PERIOD_SEC)



#!/usr/bin/env python3
import logging
import sys
from argparse import ArgumentParser
from typing import Callable, List, NoReturn, Optional, Union

import backoff
import requests

logger = logging.getLogger(__name__)

handler = logging.StreamHandler(sys.stderr)
handler.setLevel(logging.DEBUG)
no_color_formatter = logging.Formatter("[%(levelname)s] %(message)s")

color_formatter: Union[logging.Formatter, "colorlog.ColoredFormatter"]
try:
    import colorlog  # This is an optional dependency

    color_formatter = colorlog.ColoredFormatter(
        "%(log_color)s%(levelname)s %(message)s"
    )
except ImportError:  # pragma: no cover
    # Mocking an ImportError in unit tests looks tricky
    color_formatter = no_color_formatter

handler.setFormatter(color_formatter)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)


def make_wait_for_service_ready(
    max_retry_seconds: float, retry_delay_sec: float
) -> Callable[[str], NoReturn]:
    exceptions = (
        requests.exceptions.ConnectionError,
        requests.exceptions.RequestException,
        requests.exceptions.JSONDecodeError,
        RuntimeError,
    )

    @backoff.on_exception(
        backoff.constant,
        exceptions,
        logger=logger,
        jitter=None,
        max_time=max_retry_seconds,
        interval=retry_delay_sec,
    )
    def wait_for_service_ready(url: str) -> NoReturn:
        response = requests.get(url)
        response.raise_for_status()
        logger.debug("%s (status %s): %s", url, response.status_code, response.text)

        ready_status = response.json()

        if ready_status["ready"] is not True:
            raise RuntimeError(f"Service is not ready: {ready_status}")

        logger.info("Service is ready :)")
        sys.exit(0)

    def wait_for_service_ready_wrapped(url: str) -> NoReturn:
        try:
            wait_for_service_ready(url)
        except exceptions:
            logger.critical("Service did not become ready", exc_info=True)
            sys.exit(2)

    return wait_for_service_ready_wrapped


def main(argv: Optional[List[str]] = None) -> NoReturn:
    if argv is None:  # pragma: no cover
        argv = sys.argv[1:]
    argparser = ArgumentParser(
        description="Wait for a given service to mark itself ready"
    )
    argparser.add_argument("--ready-check-url", "--url", "-u", required=True)
    argparser.add_argument("--max-retry-seconds", type=int, default=7)
    argparser.add_argument("--retry-delay_sec", type=float, default=1 / 3)
    argparser.add_argument("--quiet", action="store_true")
    argparser.add_argument("--no-color", action="store_true")

    args = argparser.parse_args(argv)

    if args.quiet:
        logger.setLevel(logging.INFO)

    if args.no_color:
        handler.setFormatter(no_color_formatter)

    wait_for_service_ready = make_wait_for_service_ready(
        args.max_retry_seconds, args.retry_delay_sec
    )

    wait_for_service_ready(args.ready_check_url)


if __name__ == "__main__":  # pragma: no cover
    try:
        main()
    except KeyboardInterrupt:
        logger.info("Goodbye")
        sys.exit(0)



import abc
import json
from typing import TypedDict

from starlette.types import Receive, Scope, Send


# Necessary because mypy does not accept "TypedDict" as a type
# error: Variable "typing.TypedDict" is not valid as a type
class AnyTypedDict(TypedDict):
    pass


class ServerBase:
    @property
    @abc.abstractmethod
    def _status(self) -> int:
        pass  # pragma: no cover

    @property
    @abc.abstractmethod
    def _response(self) -> AnyTypedDict:
        pass  # pragma: no cover

    async def asgi_app(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI app which returns this class' status code and a JSON response.

        Based on the Prometheus server from the client library.
        """
        assert scope.get("type") == "http"

        status = self._status
        header = ("Content-Type", "application/json")
        output = json.dumps(self._response)

        payload = await receive()
        if payload.get("type") == "http.request":
            await send(
                {
                    "type": "http.response.start",
                    "status": status,
                    "headers": [tuple(x.encode("utf8") for x in header)],
                }
            )
            await send({"type": "http.response.body", "body": output.encode("utf8")})



import json
import logging
from typing import Dict, TypedDict

from powerflex_monitoring.server_base import ServerBase

logger = logging.getLogger(__name__)

DEFAULT_READINESS_IDENTIFIER = "default"
METRIC_NAME = "service_readiness"

try:
    from prometheus_client import Gauge

    METRIC = Gauge(
        METRIC_NAME,
        "1 if the service is marked 'ready', 0 otherwise",
        labelnames=["cause", "ready_causes", "not_ready_causes"],
    )
except ImportError:
    logger.info(
        f"Prometheus metric {METRIC_NAME} for indicating service readiness is disabled"
    )

ReadinessIdentifier = str


class ReadinessState(TypedDict):
    ready: bool
    cause: str


class ReadyCheck(ServerBase):
    def __init__(
        self,
        initial_state: bool = False,
        initial_cause: str = "Service is initializing",
    ) -> None:
        self._is_ready = initial_state
        self._cause = initial_cause
        self._ready_causes: Dict[ReadinessIdentifier, str] = {}
        self._not_ready_causes: Dict[ReadinessIdentifier, str] = {}
        self.supervisor_ready_check: Dict[ReadinessIdentifier, ReadinessState] = {}
        self._update_metric()

    @property
    def _status(self) -> int:
        return 200 if self._is_ready else 503

    @property
    def _response(self) -> ReadinessState:
        return {"ready": self._is_ready, "cause": self._cause}

    @property
    def is_ready(self) -> bool:
        return self._is_ready

    def _update_metric(self) -> None:
        METRIC.clear()
        METRIC.labels(
            **{
                "cause": self._cause,
                "ready_causes": json.dumps(self._ready_causes),
                "not_ready_causes": json.dumps(self._not_ready_causes),
            }
        ).set(1 if self._is_ready else 0)

    @staticmethod
    def _aggregate_cause(causes: Dict[ReadinessIdentifier, str]) -> str:
        return " - ".join(
            [
                f"{cause}(readiness_identifier={readiness_identifier})"
                for readiness_identifier, cause in causes.items()
            ]
        )

    def set_ready(
        self,
        ready: bool,
        readiness_identifier: ReadinessIdentifier = DEFAULT_READINESS_IDENTIFIER,
        cause: str = "Unknown",
    ) -> None:
        if (
            self.supervisor_ready_check.get(readiness_identifier) is not None
            and self.supervisor_ready_check[readiness_identifier]["ready"] == ready
        ):
            return

        self.supervisor_ready_check[readiness_identifier] = {
            "ready": ready,
            "cause": cause,
        }

        # Log whenever the readiness state may change
        log_level = logging.INFO if ready else logging.WARNING
        logger.log(
            log_level,
            "Readiness Identifier %s notifies the service as %s due to %s",
            readiness_identifier,
            "ready" if ready else "not ready",
            cause,
        )
        self._cause = ""

        # Find which causes are making the service not ready or ready
        self._not_ready_causes = {}
        self._ready_causes = {}

        # Check each readiness_identifier to see if any is unready
        # The service is ready if only all of the readiness_identifier
        for key, value in self.supervisor_ready_check.items():
            if not value["ready"]:
                self._not_ready_causes[key] = value["cause"]
            else:
                self._ready_causes[key] = value["cause"]

        # If any of the readiness_identifiers is not ready, the service is not ready
        if len(self._not_ready_causes) > 0:
            self._is_ready = False
            self._cause = self._aggregate_cause(self._not_ready_causes)
            self._update_metric()

            logger.warning(
                "Service has been marked as not ready cause(s)=%s.",
                self._cause,
                stack_info=True,
                extra={
                    "ready": self._is_ready,
                    "cause": self._cause,
                    "causes": self._not_ready_causes,
                },
            )
            return

        self._is_ready = True
        # Report all readiness causes
        self._cause = self._aggregate_cause(self._ready_causes)
        self._update_metric()
        logger.info(
            "Service has been marked as ready cause(s)=%s.",
            self._cause,
            extra={
                "ready": self._is_ready,
                "cause": self._cause,
                "causes": self._ready_causes,
            },
        )



import inspect
from typing import Awaitable, Callable, Optional

from prometheus_client import generate_latest
from starlette.types import Receive, Scope, Send

PROMETHEUS_CONTENT_TYPE_LATEST = "text/plain; version=0.0.4; charset=utf-8"


async def prometheus_asgi_app(scope: Scope, receive: Receive, send: Send) -> None:
    """ASGI app which returns the prometheus metrics as plain text via HTTP."""
    assert scope.get("type") == "http"

    headers = ("Content-Type", PROMETHEUS_CONTENT_TYPE_LATEST)

    output = generate_latest()

    payload = await receive()
    if payload.get("type") == "http.request":
        await send(
            {
                "type": "http.response.start",
                "status": 200,
                "headers": [tuple(header.encode("utf8") for header in headers)],
            }
        )
        await send({"type": "http.response.body", "body": output})


class PrometheusASGIAppWithCallback:
    def __init__(
        self,
        after_metrics_scraped: Optional[Callable[[], Optional[Awaitable[None]]]] = None,
    ) -> None:
        self.after_metrics_scraped = after_metrics_scraped

    def set_after_metrics_scraped(
        self, after_metrics_scraped: Callable[[], Optional[Awaitable[None]]]
    ) -> None:
        self.after_metrics_scraped = after_metrics_scraped

    async def asgi_app(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI app which returns the prometheus metrics as plain text."""
        await prometheus_asgi_app(scope, receive, send)

        if self.after_metrics_scraped is not None:
            result = self.after_metrics_scraped()
            if inspect.isawaitable(result):
                await result



import logging
from typing import Type

import uvicorn  # type: ignore
from starlette.routing import Mount, Router

from powerflex_monitoring.health_check import HealthCheck
from powerflex_monitoring.prometheus_metrics import PrometheusASGIAppWithCallback
from powerflex_monitoring.ready_check import ReadyCheck

logger = logging.getLogger(__name__)


class MonitoringServer:
    def __init__(
        self,
        port: int,
        health_check_factory: Type[HealthCheck] = HealthCheck,
        ready_check_factory: Type[ReadyCheck] = ReadyCheck,
        metrics_app_factory: Type[
            PrometheusASGIAppWithCallback
        ] = PrometheusASGIAppWithCallback,
    ):
        self.port = port
        self.health_check = health_check_factory()
        self.ready_check = ready_check_factory()
        self.metrics_app = metrics_app_factory()

        health_check_name = "health"
        ready_check_name = "ready"
        metrics_name = "metrics"
        self.route_names = [health_check_name, ready_check_name, metrics_name]

        self.app = Router(
            routes=[
                Mount(
                    "/monitoring",
                    routes=[
                        Mount(
                            "/health",
                            app=self.health_check.asgi_app,
                            name=health_check_name,
                        ),
                        Mount(
                            "/ready",
                            app=self.ready_check.asgi_app,
                            name=ready_check_name,
                        ),
                        Mount(
                            "/metrics",
                            app=self.metrics_app.asgi_app,
                            name=metrics_name,
                        ),
                    ],
                )
            ]
        )

    def start(self) -> None:
        for name in self.route_names:
            logger.info(
                "Listening at HTTP route at port %s: %s",
                self.port,
                self.app.url_path_for(name, path="/"),
            )

        logger.info("🚀 Starting monitoring server at port %s", self.port)
        uvicorn.run(
            # For some reason, the types that come with starlette don't line up
            # with the type hints for uvicorn.
            self.app,  # type: ignore
            host="0.0.0.0",
            port=self.port,
            log_level="warning",
        )



import logging
from typing import TypedDict

from powerflex_monitoring.server_base import ServerBase

logger = logging.getLogger(__name__)

METRIC_NAME = "service_health"

try:
    from prometheus_client import Gauge

    METRIC = Gauge(
        METRIC_NAME,
        "1 if the service is healthy, 0 if it is unhealthy",
        labelnames=["cause"],
    )
except ImportError:
    logger.info(
        f"Prometheus metric {METRIC_NAME} for indicating service readiness is disabled"
    )


class HealthCheckResponse(TypedDict):
    healthy: bool
    cause: str


class HealthCheck(ServerBase):
    def __init__(self) -> None:
        self._is_healthy = True
        self._cause = "Initial state"

        self._update_metric()

    @property
    def _status(self) -> int:
        return 200 if self._is_healthy else 503

    @property
    def _response(self) -> HealthCheckResponse:
        return {"healthy": self._is_healthy, "cause": self._cause}

    def _update_metric(self) -> None:
        METRIC.clear()
        METRIC.labels(
            **{
                "cause": self._cause,
            }
        ).set(1 if self._is_healthy else 0)

    def make_unhealthy(self, cause: str = "Unknown") -> None:
        logger.error(
            "Service has been marked as unhealthy. Cause: %s", cause, stack_info=True
        )
        self._is_healthy = False
        self._cause = cause

        self._update_metric()



from __future__ import annotations

import asyncio
import logging
import socket
import time
from collections import defaultdict
from typing import (
    Any,
    DefaultDict,
    Dict,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseModel, Field
from redis import Redis
from redis.asyncio import Redis as AsyncRedis

from powerflex_monitoring.format_exception import format_exception
from powerflex_monitoring.monitoring import MonitoringServer

BaseSettings: Type[BaseModel]
try:
    from pydantic_settings import BaseSettings
except ImportError:
    # pylint: disable=ungrouped-imports
    from pydantic import BaseSettings  # type: ignore

logger = logging.getLogger(__name__)
HOSTNAME_TEMPLATE_VAR = "{hostname}"


StrType = TypeVar("StrType", bound=Union[str, bytes])


class RedisHealthCheckerConfig(BaseSettings):  # type: ignore
    REDIS_HEALTH_CHECK_KEY_TEMPLATE: str = Field(
        description="Redis key on which this service will SET and GET to periodically. "
        f"The string {HOSTNAME_TEMPLATE_VAR} is replaced with the service's hostname. "
        "The hostname is used in this Redis key to ensure it is unique across all containers. "
        "In docker-compose and kubernetes, the hostname is typically container-specific. ",
        default=f"{HOSTNAME_TEMPLATE_VAR}.redis-health-check_key",
    )

    REDIS_HEALTH_CHECK_UNHEALTHY_TIMEOUT_SEC: float = Field(
        description="Mark the service as unhealthy after determining that Redis is unhealthy for this period of time. "
        "Recommend making it less than REDIS_HEALTH_CHECK_REQUEST_PERIOD_SEC.",
        default=60 * 4,
    )

    REDIS_HEALTH_CHECK_REQUEST_PERIOD_SEC: float = Field(
        description="How often to periodically SET and GET data to test the Redis connection.",
        default=15,
    )
    REDIS_MASTER_SET: str = Field(
        description="Redis master set name.",
        default="mymaster",
    )


class RedisHealthChecker:
    redis_conn: Union[  # type: ignore
        Redis,
        AsyncRedis,
    ]

    def __init__(
        self,
        redis_conn: Union[Redis[StrType], AsyncRedis[StrType]],
        monitoring_server: MonitoringServer,
    ):
        self.config = RedisHealthCheckerConfig(
            # type:ignore
        )

        logger.info(
            "Configuration for Redis health check", extra={"config": self.config.dict()}
        )

        self.redis_health_check_key = (
            # pylint: disable=no-member
            self.config.REDIS_HEALTH_CHECK_KEY_TEMPLATE.replace(
                HOSTNAME_TEMPLATE_VAR,
                socket.gethostname(),
            )
        )
        self.redis_conn = redis_conn
        self.monitoring_server = monitoring_server
        self.time_since_redis_unhealthy: DefaultDict[str, Optional[float]] = (
            defaultdict(lambda: None)
        )

    def on_unhealthy(self, health_check_name: str, cause: str) -> None:
        """Check if a Redis health check has been unhealthy for too long.

        If so, then mark the service as unhealthy.
        """
        self.on_redis_health_status(
            healthy=False, health_check_name=health_check_name, cause=cause
        )

        time_since = self.time_since_redis_unhealthy[health_check_name]
        unhealthy_for_too_long = (
            time_since is not None
            and time.time() - time_since
            >= self.config.REDIS_HEALTH_CHECK_UNHEALTHY_TIMEOUT_SEC
        )

        if unhealthy_for_too_long:
            logger.error(
                f"Redis health check {health_check_name} has been unhealthy for too long. Marking service as unhealthy."
            )
            self.monitoring_server.health_check.make_unhealthy(cause)

    def on_redis_health_status(
        self, healthy: bool, health_check_name: str, cause: str = "Redis is healthy"
    ) -> None:
        """Call this whenever a result for a Redis health check is found."""
        if healthy is True:
            logger.debug(
                "Redis health check %s indicates Redis is healthy", health_check_name
            )
            self.time_since_redis_unhealthy[health_check_name] = None

        if healthy is False:
            self.monitoring_server.ready_check.set_ready(
                ready=False, cause=cause, readiness_identifier="redis-identifier"
            )

            if self.time_since_redis_unhealthy[health_check_name] is None:
                logger.info(
                    "Redis health check %s indicates Redis is currently unhealthy",
                    health_check_name,
                    extra={"cause": cause},
                )
                self.time_since_redis_unhealthy[health_check_name] = time.time()

        all_health_checks_pass = all(
            check_failure_time is None
            for check_failure_time in self.time_since_redis_unhealthy.values()
        )
        if not self.monitoring_server.ready_check.is_ready and all_health_checks_pass:
            self.monitoring_server.ready_check.set_ready(
                ready=True,
                cause="Redis is healthy again",
                readiness_identifier="redis-identifier",
            )

    async def check_write_read_redis_service(  # pylint: disable=too-many-branches
        self,
    ) -> None:
        """Make a Redis write-read operation with a unique key to check Redis health."""
        health_check_name = self.check_write_read_redis_service.__name__
        test_value = str(time.time())
        result: Optional[bool] = False
        redis_value: Optional[str] = None

        try:
            if isinstance(self.redis_conn, AsyncRedis):
                result = await self.redis_conn.set(
                    self.redis_health_check_key, test_value
                )
            else:
                result = self.redis_conn.set(self.redis_health_check_key, test_value)

        except ConnectionError as exc:
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause="Redis set to test writing a document failed. Exception: "
                + format_exception(exc),
            )
            return
        except Exception as exc:  # pylint: disable=broad-except
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause="Unexpected exception when testing Redis SET operation. Exception: "
                + format_exception(exc),
            )
            return

        if result:
            try:
                if isinstance(self.redis_conn, AsyncRedis):
                    _redis_value = await self.redis_conn.get(
                        self.redis_health_check_key
                    )
                else:
                    _redis_value = self.redis_conn.get(self.redis_health_check_key)
                if _redis_value is not None:
                    if isinstance(_redis_value, bytes):
                        redis_value = _redis_value.decode("utf-8")
                    else:
                        redis_value = str(_redis_value)
            except ConnectionError as exc:
                self.on_unhealthy(
                    health_check_name=health_check_name,
                    cause="Redis get to test reading a document failed. Exception: "
                    + format_exception(exc),
                )
                return
            except Exception as exc:  # pylint: disable=broad-except
                self.on_unhealthy(
                    health_check_name=health_check_name,
                    cause="Unexpected exception when testing Redis GET operation. Exception: "
                    + format_exception(exc),
                )
                return
            if redis_value == test_value:
                self.on_redis_health_status(
                    healthy=True, health_check_name=health_check_name
                )
            else:
                self.on_unhealthy(
                    health_check_name=health_check_name,
                    cause="Redis GET operation returned an unexpected value."
                    f"Expected: {test_value}. Actual: {redis_value}",
                )

    async def check_redis_service_health(self) -> None:
        """Check the Redis service's health to determine if it is healthy using the ping command."""
        health_check_name = self.check_redis_service_health.__name__
        try:
            if isinstance(self.redis_conn, AsyncRedis):
                is_connected = await self.redis_conn.ping()
            else:
                is_connected = self.redis_conn.ping()
        except ConnectionError as exc:
            connection_kwargs = cast(Dict[str, Any], self.redis_conn.get_connection_kwargs())  # type: ignore[union-attr]
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause="Could not connect to Redis service health check. "
                + f"Host: {connection_kwargs.get('host', 'unknown')}, "
                + f"Port: {connection_kwargs.get('port', 'unknown')}. Exception: "
                + format_exception(exc),
            )
            return
        except Exception as exc:  # pylint: disable=broad-except
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause="Unexpected exception when trying to connect to Redis. Exception: "
                + format_exception(exc),
            )
            return

        if is_connected:
            self.on_redis_health_status(
                healthy=True, health_check_name=health_check_name
            )
        else:
            self.on_unhealthy(
                health_check_name=health_check_name,
                cause="Redis service health check returned not connected",
            )

    async def periodically_check_redis_health(self) -> NoReturn:
        """Periodically check the Redis health by running the following checks.

        1. Check if a Redis write-read operation works.
        2. Check if the Redis service's health check is bad

        If any health check keeps failing for longer than
        REDIS_HEALTH_CHECK_UNHEALTHY_TIMEOUT_SEC,
        then mark the service as unhealth.

        If any health check fails at all, mark the service as not ready.

        Mark the service as ready if all health checks pass.
        """
        logger.info("Starting Redis health checkers")
        while True:
            await self.check_write_read_redis_service()
            await self.check_redis_service_health()
            await asyncio.sleep(self.config.REDIS_HEALTH_CHECK_REQUEST_PERIOD_SEC)



"""Developer diagnostic for historical station information.

Write-only and relies on external tooling such as the
nats CLI for viewing data e.g.

List recorded stations:
  $ nats kv ls evse

List history of a given key:
  $ nats kv history evse 0001-50-01-08

Get a historical value in full for a given station and revision (from
the history command):
  $ nats kv get evse 0001-50-01-08 --raw --revision 10

"""

from __future__ import annotations

import logging
import re
from typing import Optional, Pattern

from nats.aio.client import Client as NATSClient
from nats.js.kv import KeyValue

logger = logging.getLogger(__name__)


class DiagnosticRecorder:
    def __init__(self) -> None:
        self.key_value: Optional[KeyValue] = None
        self.key_validator: Pattern[str] = re.compile(
            r"\A(?!_kv)(?!\.)(?!.*\.\Z)[-/_=\.a-zA-Z0-9]+\Z"
        )
        self.history = 32
        self.bucket = "evse"

    @classmethod
    async def create(cls, nats_conn: NATSClient) -> DiagnosticRecorder:
        instance = DiagnosticRecorder()
        jetstream = nats_conn.jetstream()
        try:
            instance.key_value = await jetstream.key_value(bucket=instance.bucket)
        except Exception:  # pylint: disable=broad-except
            instance.key_value = await jetstream.create_key_value(
                bucket=instance.bucket, history=instance.history
            )
        return instance

    async def record(self, key: str, value: bytes) -> None:
        try:
            self.validate_key(key)
            await self.key_value.put(key, value)  # type: ignore
        except Exception:  # pylint: disable=broad-except
            logger.warning("failed to record diagnostic for %s", key, exc_info=True)

    def validate_key(self, key: str) -> None:
        """Validate implicit NATS requirement.

        Key-Value has an unchecked requirement on the format of keys
        which, when unmet, can result in confusing errors
        (e.g. timeouts). The regular expression is derived from NATS
        ADR-8

        """
        assert self.key_validator.match(key) is not None
